"""Top level tet module for storing fixtures."""

import pytest
from gcloud_storage_emulator.server import Server as GCloudStorageMockServer
from google.cloud import storage


@pytest.fixture(scope="function")
def google_cloud_storage():
    """Fixture to start the gcloud storage emulator."""
    host = "localhost"
    port = 4443
    in_memory = True
    with pytest.MonkeyPatch.context() as m:
        m.setenv("STORAGE_EMULATOR_HOST", f"http://{host}:{port}")
        emulator = GCloudStorageMockServer(host=host, port=port, in_memory=in_memory)
        emulator.start()
        storage.Client().create_bucket("test")  # Create a test bucket
        yield emulator
        emulator.wipe()
        emulator.stop()


@pytest.mark.usefixtures("google_cloud_storage")
@pytest.fixture(scope="function")
def staging_bucket():
    """Fixture to create a staging bucket."""
    bucket_name = "staging"
    client = storage.Client()
    client.create_bucket(bucket_name)
    return bucket_name


@pytest.mark.usefixtures("google_cloud_storage")
@pytest.fixture(scope="function")
def gwas_catalog_bucket():
    """Fixture to create a gwas catalog bucket."""
    bucket_name = "gwas_catalog"
    client = storage.Client()
    client.create_bucket(bucket_name)
    return bucket_name



"""Test cases for the update_gwas_curation_metadata command."""

import logging
from ftplib import FTP
from pathlib import Path

import gentroutils
import pytest
from click.testing import CliRunner
from gentroutils import cli
from google.cloud import storage


def gwas_catalog_ftp_heartbeet() -> bool:
    """Check if GWAS Catalog FTP server is up and running."""
    try:
        with FTP("ftp.ebi.ac.uka") as ftp:
            ftp.login()
            ftp.voidcmd("NOOP")
            return True
    except Exception:
        return False


@pytest.mark.integration_test
def test_run_update_gwas_curation_metadata_help():
    """Test command with --help flag."""
    runner = CliRunner()
    result = runner.invoke(cli, ["update-gwas-curation-metadata", "--help"])
    assert result.exit_code == 0


@pytest.mark.integration_test
def test_run_update_gwas_curation_metadata_exceed_connection(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
):
    """Test command with --dry-run flag when connection limit is exceeded."""
    monkeypatch.setattr(
        gentroutils.commands.update_gwas_curation_metadata,
        "MAX_CONCURRENT_CONNECTIONS",
        0,
    )
    runner = CliRunner()
    result = runner.invoke(cli, ["--dry-run", "update-gwas-curation-metadata"])
    for record in caplog.records:
        if record.levelname == "ERROR":
            assert (
                "File transfer limit exceeded! Max 0 connections allowed" in caplog.text
            )
    assert result.exit_code == 1


@pytest.mark.usefixtures("google_cloud_storage", "staging_bucket")
@pytest.mark.skipif(
    gwas_catalog_ftp_heartbeet(), reason="GWAS Catalog FTP not accessible"
)
@pytest.mark.integration_test
def test_run_update_gwas_curation_metadata_no_dry_run(caplog: pytest.LogCaptureFixture):
    """Test command save from gwas catalog ftp to local gcs mock server."""
    caplog.set_level(logging.DEBUG)
    runner = CliRunner()
    _in = (
        "ftp://ftp.ebi.ac.uk/pub/databases/gwas/summary_statistics/harmonised_list.txt"
    )
    _out = "gs://staging/harmonised_list.txt"
    result = runner.invoke(cli, ["update-gwas-curation-metadata", "-f", _in, _out])
    assert result.exit_code == 0
    client = storage.Client()
    blobs = client.bucket("staging").list_blobs()
    assert next(blobs).name == "harmonised_list.txt"
    for record in caplog.records:
        if record.levelname == "INFO" and "Retrieving data" in record.message:
            assert f"Retrieving data from: {_in}." in record.message
        if record.levelname == "INFO" and "Uploading data" in record.message:
            assert f"Uploading data to: {_out}." in record.message
        if record.levelname == "INFO" and "Diseases" in record.message:
            assert "Diseases were mapped to v" in record.message
        if record.levelname == "INFO" and "EFO version" in record.message:
            assert "EFO version: v" in record.message


@pytest.mark.usefixtures("google_cloud_storage", "staging_bucket")
@pytest.mark.integration_test
def test_run_update_gwas_curation_metadata_fail_to_fetch_release_info(
    caplog: pytest.LogCaptureFixture,
):
    """Test command to see if the release info is correctly fetched."""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "update-gwas-curation-metadata",
            "-f",
            "ftp://ftp.ebi.ac.uk/pub/databases/gwas/summary_statistics/harmonised_list.txt",
            "gs://staging/harmonised_list.txt",
            "-g"
            "http://localhost:4443",  # incorrect link to the gwas catalog api stats
        ],
    )
    assert result.exit_code == 1
    for record in caplog.records:
        if record.levelname == "ERROR":
            assert "Failed to fetch release info" in record.message


@pytest.mark.usefixtures("google_cloud_storage", "staging_bucket")
@pytest.mark.parametrize(
    "link",
    [
        pytest.param(
            "ftp://ftp.ebi.ac.uk/pub/databases/gwas/summary_statistics/harmonised_list.txt",
            id="ftp",
        ),
        pytest.param(
            "https://raw.githubusercontent.com/opentargets/curation/master/genetics/GWAS_Catalog_study_curation.tsv",
            id="http",
        ),
    ],
)
@pytest.mark.skipif(
    gwas_catalog_ftp_heartbeet(), reason="GWAS Catalog FTP not accessible"
)
@pytest.mark.integration_test
def test_run_update_gwas_curation_metadata_with_dry_run_does_not_produce_blob(
    link: str,
):
    """Test command to see if the release info is correctly fetched."""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--dry-run",
            "update-gwas-curation-metadata",
            "-f",
            link,
            "gs://staging/harmonised_list.txt",
        ],
    )
    assert result.exit_code == 0
    client = storage.Client()
    blobs = client.bucket("staging").list_blobs()
    assert not list(blobs)


@pytest.mark.usefixtures("google_cloud_storage", "staging_bucket")
@pytest.mark.skipif(
    gwas_catalog_ftp_heartbeet(), reason="GWAS Catalog FTP not accessible"
)
@pytest.mark.integration_test
def test_run_update_gwas_curation_metadata_transfer_from_http_to_gcp():
    """Test command to see if the release info is correctly fetched."""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "update-gwas-curation-metadata",
            "-f",
            "https://raw.githubusercontent.com/opentargets/curation/master/genetics/GWAS_Catalog_study_curation.tsv",
            "gs://staging/curation-metadata.tsv",
        ],
    )
    assert result.exit_code == 0
    client = storage.Client()
    blobs = client.bucket("staging").list_blobs()
    assert next(blobs).name == "curation-metadata.tsv"


@pytest.mark.usefixtures("google_cloud_storage", "staging_bucket")
@pytest.mark.skipif(
    gwas_catalog_ftp_heartbeet(), reason="GWAS Catalog FTP not accessible"
)
@pytest.mark.integration_test
def test_run_update_gwas_curation_metadata_preserve_logs_locally(tmp_path: Path):
    """Test command dumps logs to a local file."""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--log-file",
            str(tmp_path / "gentroutils.log"),
            "update-gwas-curation-metadata",
            "-f",
            "ftp://ftp.ebi.ac.uk/pub/databases/gwas/summary_statistics/harmonised_list.txt",
            "gs://staging/harmonised_list.txt",
        ],
    )
    assert result.exit_code == 0
    client = storage.Client()
    blobs = client.bucket("staging").list_blobs()
    assert next(blobs).name == "harmonised_list.txt"
    assert (tmp_path / "gentroutils.log").exists()
    Path(tmp_path / "gentroutils.log").unlink()


@pytest.mark.usefixtures("google_cloud_storage", "staging_bucket")
@pytest.mark.skipif(
    gwas_catalog_ftp_heartbeet(), reason="GWAS Catalog FTP not accessible"
)
@pytest.mark.integration_test
def test_run_update_gwas_curation_metadata_preserve_logs_in_gcs():
    """Test command dumps logs to a gcs file."""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--log-file",
            "gs://staging/gentroutils.log",
            "update-gwas-curation-metadata",
            "-f",
            "ftp://ftp.ebi.ac.uk/pub/databases/gwas/summary_statistics/harmonised_list.txt",
            "gs://staging/harmonised_list.txt",
        ],
    )
    assert result.exit_code == 0
    client = storage.Client()
    blobs = client.bucket("staging").list_blobs()
    for blob in blobs:
        assert blob.name == "harmonised_list.txt" or blob.name == "gentroutils.log"



"""Ütility functions for the CLI."""

import asyncio
import logging
import sys
import time
from functools import wraps
from pathlib import Path
from tempfile import NamedTemporaryFile
from urllib.parse import urlparse

import click
from google.cloud import storage

logger = logging.getLogger("gentroutils")


def set_log_file(ctx: click.Context, param: click.Option, log_file: str) -> str:
    """Set logging file based on provided `log-file` flag.

    This is a callback function called by the click.Option [--log-file] flag.
    In case of the `log_file` being path to the GCP bucket the returned value
    will be the local temporary file path. both log file paths (remote and local)
    will be stored in the click context object for further reference at the end of the CLI run.


    Args:
        ctx (click.Context): click context
        param (click.Option): click option
        log_file (str): log file path

    Raises:
        click.BadParameter: If the log file is a directory or the URI scheme is not GCS.

    Returns:
        str: log file path
    """
    ctx.ensure_object(dict)
    if not log_file:
        return ""
    logger.info("Extracting log file from the %s", param)
    upload_to_gcp = False

    if "://" in log_file:
        upload_to_gcp = True
    ctx.obj["upload_to_gcp"] = upload_to_gcp

    if upload_to_gcp:
        parsed_uri = urlparse(log_file)
        if parsed_uri.scheme != "gs":
            raise click.BadParameter("Only GCS is supported for logging upload")
        tmp_file = NamedTemporaryFile(delete=False)
        logger.info("Logging to temporary file %s", tmp_file.name)
        handler = logging.FileHandler(tmp_file.name)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        ctx.obj["local_log_file"] = tmp_file.name
        ctx.obj["local_log_file_obj"] = tmp_file
        ctx.obj["gcp_log_file"] = log_file
        return tmp_file.name

    else:
        local_file = Path(log_file)
        if local_file.exists() and local_file.is_dir():
            raise click.BadParameter("Log file is a directory")
        if local_file.exists() and local_file.is_file():
            local_file.unlink()
        if not local_file.exists():
            local_file.parent.mkdir(parents=True, exist_ok=True)
            local_file.touch()
        logger.info("Logging to %s", local_file)
        handler = logging.FileHandler(local_file)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        return str(local_file)


def teardown_cli(ctx: click.Context) -> None:
    """Teardown the gentroutils cli.

    This function is used to as a teardown function for the CLI.
    This will upload the log file to the GCP bucket if the `upload_to_gcp` flag is set in the context object.

    Args:
        ctx (click.Context): click context
    """
    if "upload_to_gcp" in ctx.obj and ctx.obj["upload_to_gcp"]:
        gcp_file = ctx.obj["gcp_log_file"]
        local_file = ctx.obj["local_log_file"]
        with open(local_file, "r") as f:
            content = f.read()
        try:
            client = storage.Client()
            bucket_name = urlparse(gcp_file).netloc
            bucket = client.bucket(bucket_name=bucket_name)
            file_name = urlparse(gcp_file).path.lstrip("/")
            blob = bucket.blob(file_name)
            logger.info("Uploading %s to %s", local_file, gcp_file)
            if ctx.obj["dry_run"]:
                logger.info("Dry run, skipping the upload of the log file")
            else:
                blob.upload_from_string(content)
                ctx.obj["local_log_file_obj"].close()
        except Exception as e:
            msg = f"Failed to upload log file to GCP {e}"
            logger.error(click.style(msg, fg="red"))
    logger.info(
        "Finished, elapsed time %s seconds", time.time() - ctx.obj["execution_start"]
    )


def set_log_lvl(_: click.Context, param: click.Option, value: int) -> int:
    """Set logging level based on the number of provided `v` flags.

    This is a callback function called by the click.Option [-v] flag.
    For example
    `-vv` - DEBUG
    `-v`  - INFO
    `no flag - ERROR

    Args:
        param (click.Option): click option
        value (int): logging level

    Returns:
        int: logging level
    """
    logger.info("Extracting log level from the %s", param)
    log_lvls = {0: logging.ERROR, 1: logging.INFO, 2: logging.DEBUG}
    log_lvl = log_lvls.get(value, logging.DEBUG)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    handler.setLevel(log_lvl)
    logger.addHandler(handler)
    return log_lvl


def coro(f):
    """Corutine wrapper for synchronous functions."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        """Wrapper around the synchronous function."""
        return asyncio.run(f(*args, **kwargs))

    return wrapper


__all__ = ["set_log_file", "set_log_lvl", "coro", "logger", "teardown_cli"]



"""Update gwas catalog metadata."""

from __future__ import annotations

import asyncio
import io
import logging
import re
import sys
from ftplib import FTP
from urllib.parse import ParseResult, urlparse

import click
import requests
from google.cloud import storage

from gentroutils.commands.utils import coro

logger = logging.getLogger("gentroutils")
MAX_CONCURRENT_CONNECTIONS = 10
CURATED_INPUTS = (
    (
        "ftp://ftp.ebi.ac.uk/pub/databases/gwas/releases/latest/gwas-catalog-associations_ontology-annotated.tsv",
        "gs://gwas_catalog_inputs/gwas_catalog_associations_ontology_annotated.tsv",
    ),
    (
        "ftp://ftp.ebi.ac.uk/pub/databases/gwas/releases/latest/gwas-catalog-download-studies-v1.0.3.1.txt",
        "gs://gwas_catalog_inputs/gwas_catalog_download_studies.tsv",
    ),
    (
        "ftp://ftp.ebi.ac.uk/pub/databases/gwas/releases/latest/gwas-catalog-download-ancestries-v1.0.3.1.txt",
        "gs://gwas_catalog_inputs/gwas_catalog_download_ancestries.tsv",
    ),
)


@click.command(name="update-gwas-curation-metadata")
@click.option(
    "--file-to-transfer",
    "-f",
    metavar="<ftp_file|http(s) file> <gcp_file>",
    type=(str, str),
    multiple=True,
    default=CURATED_INPUTS,
)
@click.option(
    "--gwas-catalog-release-info-url",
    "-g",
    metavar="<url>",
    default="https://www.ebi.ac.uk/gwas/api/search/stats",
    type=click.STRING,
)
@click.pass_context
@coro
async def update_gwas_curation_metadata_command(
    ctx: click.Context,
    file_to_transfer: list[tuple[str, str]],
    gwas_catalog_release_info_url: str,
) -> None:
    """Update GWAS Catalog metadata directly to cloud bucket.

    \b
    This is the script to fetch the latest GWAS Catalog data files that include:
    - [x] gwas-catalog-associations_ontology-annotated.tsv - list of associations with ontology annotations by GWAS Catalog
    - [x] gwas-catalog-download-studies-v1.0.3.1.txt - list of published studies by GWAS Catalog
    - [x] gwas-catalog-download-ancestries-v1.0.3.1.txt - list of published studies by GWAS Catalog

    \b
    By default all GWAS Catalog data files are uploaded from GWAS Catalog FTP server to Open Targets GCP bucket.
    The script also captures the latest release metadata from GWAS Catalog release info url.
    One can overwrite this script to sync data files from FTP or HTTP(s) to GCP bucket. The example usage is as follows:

    \b
    gentroutils --log-file gs://gwas_catalog_data/curated_inputs/log.txt update-gwas-curation-metadata \\
    -f ftp://ftp.ebi.ac.uk/pub/databases/gwas/releases/latest/gwas-catalog-associations_ontology-annotated.tsv gs://gwas_catalog_data/gwas_catalog_associations_ontology_annotated.tsv \\
    -f ftp://ftp.ebi.ac.uk/pub/databases/gwas/releases/latest/gwas-catalog-download-studies-v1.0.3.1.txt gs://gwas_catalog_inputs/gwas_catalog_download_studies.tsv \\
    -f ftp://ftp.ebi.ac.uk/pub/databases/gwas/releases/latest/gwas-catalog-download-ancestries-v1.0.3.1.txt gs://gwas_catalog_inputs/gwas_catalog_download_ancestries.tsv \\
    -g https://www.ebi.ac.uk/gwas/api/search/stats


    To preserve the logs from this command, you can specify the log file path using `--log-file` option. The log file can point to local or GCP path.
    Currently only FTP and HTTP(s) protocols are supported for input and GCP protocol is supported for output.
    """
    # we always want to have the logs from this command uploaded to the target bucket
    logger.debug("Running gwas_curation_update step.")
    dry_run = ctx.obj["dry_run"]
    if len(file_to_transfer) > MAX_CONCURRENT_CONNECTIONS:
        logger.error(
            "File transfer limit exceeded! Max %s connections allowed.",
            MAX_CONCURRENT_CONNECTIONS,
        )
        sys.exit(1)
    uri_map = [
        {"input": urlparse(ftp_file), "output": urlparse(gcp_file)}
        for ftp_file, gcp_file in file_to_transfer
    ]
    transfer_tasks = generate_transfer_tasks(uri_map, dry_run)

    # capture latest release metadata
    with requests.get(gwas_catalog_release_info_url) as response:
        if not response.ok:
            logger.error("Failed to fetch release info.")
            sys.exit(1)
        release_info = response.json()
        for key, value in release_info.items():
            logger.debug("%s: %s", key, value)

        efo_version = release_info.get("efoversion")
        logger.info("Diseases were mapped to %s EFO release.", efo_version)
        logger.info("EFO version: %s", efo_version)
        ensembl_build = release_info.get("ensemblbuild")
        logger.info("Genes were mapped to v%s Ensembl release.", ensembl_build)

    results = await asyncio.gather(*transfer_tasks)
    if not dry_run:
        logger.info("Transferred %s files.", len(results))
    logger.info("gwas_curation_update step completed.")


def generate_transfer_tasks(
    uri_map: list[dict[str, ParseResult]], dry_run: bool
) -> list[asyncio.Task[None]]:
    """Generate transfer tasks.

    Args:
        uri_map (list[dict[str, ParseResult]]): list of transferable tasks, each should have `input` and `output` keys.
        dry_run (bool): dry run flag.

    Returns:
        list[asyncio.Task[None]]: list of asyncio tasks.
    """
    ftp_transfer_list = []
    http_transfer_list = []
    for uri in uri_map:
        if uri["input"].scheme != "ftp" and not uri["input"].scheme.startswith("http"):
            logger.error("Only FTP and HTTP(s) protocols is supported at input.")
            sys.exit(1)
        if uri["output"].scheme != "gs":
            logger.error("Only GCP protocol is supported at output.")
            sys.exit(1)
        in_server = uri["input"].netloc
        out_server = uri["output"].netloc
        in_prefix = "/".join(uri["input"].path.strip("/").split("/")[:-1])
        in_file = uri["input"].path.strip("/").split("/")[-1]
        out_prefix = "/".join(uri["output"].path[1:-1].split("/")[:-1])
        out_bucket = uri["output"].path.split("/")[-1]
        if uri["input"].scheme == "ftp":
            ftp_transfer_list.append(
                {
                    "ftp_server": in_server,
                    "ftp_prefix": in_prefix,
                    "ftp_filename": in_file,
                    "gcp_bucket": out_server,
                    "gcp_prefix": out_prefix,
                    "gcp_filename": out_bucket,
                }
            )
        if uri["input"].scheme.startswith("http"):
            http_transfer_list.append(
                {
                    "http_url": uri["input"].geturl(),
                    "gcp_bucket": out_server,
                    "gcp_prefix": out_prefix,
                    "gcp_filename": out_bucket,
                }
            )
    transfer_tasks = []
    for transfer_obj in ftp_transfer_list:
        transfer_tasks.append(
            asyncio.create_task(
                sync_from_ftp_to_gcp(
                    transfer_obj["ftp_server"],
                    transfer_obj["ftp_prefix"],
                    transfer_obj["ftp_filename"],
                    transfer_obj["gcp_bucket"],
                    transfer_obj["gcp_prefix"],
                    transfer_obj["gcp_filename"],
                    dry_run=dry_run,
                )
            )
        )

    for transfer_obj in http_transfer_list:
        transfer_tasks.append(
            asyncio.create_task(
                sync_from_http_to_gcp(
                    transfer_obj["http_url"],
                    transfer_obj["gcp_bucket"],
                    transfer_obj["gcp_prefix"],
                    transfer_obj["gcp_filename"],
                    dry_run=dry_run,
                )
            )
        )

    return transfer_tasks


async def sync_from_http_to_gcp(
    url: str, gcp_bucket: str, gcp_prefix: str, gcp_file: str, *, dry_run: bool = True
) -> None:
    """Sync file from HTTP and upload to GCP.

    This function fetches the data from the provided HTTP URL and uploads the content
    directly to provided GCP bucket blob.

    Args:
        url (str): HTTP URL to fetch the data.
        gcp_bucket (str): GCP bucket name.
        gcp_prefix (str): GCP prefix.
        gcp_file (str): GCP file name.
        dry_run (bool, optional): Dry run flag. Defaults to True.
    """
    if dry_run:
        logger.info(
            "Attempting to transfer data from %s to gs://%s/%s/%s.",
            url,
            gcp_bucket,
            gcp_prefix,
            gcp_file,
        )
        return
    logger.info("Retriving data from: %s.", url)
    response = requests.get(url)
    if not response.ok:
        logger.error("Failed to fetch data from %s.", url)
        return

    content = response.content
    bucket = storage.Client().bucket(gcp_bucket)
    gcp_path = f"{gcp_prefix}/{gcp_file}" if gcp_prefix else gcp_file

    blob = bucket.blob(gcp_path)
    logger.info("Uploading the data to: gs://%s/%s.", gcp_bucket, gcp_path)
    blob.upload_from_string(content.decode("utf-8"))


async def sync_from_ftp_to_gcp(
    ftp_server: str,
    ftp_prefix: str,
    ftp_file: str,
    gcp_bucket: str,
    gcp_prefix: str,
    gcp_file: str,
    *,
    dry_run: bool = True,
) -> None:
    """Fetch files from FTP and upload to GCP.

    This function fetches the data from the provided FTP server and uploads the content directly
    to the provided GCP bucket blob.

    Args:
        ftp_server (str): FTP server.
        ftp_prefix (str): FTP prefix.
        ftp_file (str): FTP file name.
        gcp_bucket (str): GCP bucket name.
        gcp_prefix (str): GCP prefix.
        gcp_file (str): GCP file name.
        dry_run (bool, optional): Dry run flag. Defaults to True.

    """
    if dry_run:
        logger.info(
            "Attempting to transfer data from ftp://%s/%s/%s to gs://%s/%s/%s.",
            ftp_server,
            ftp_prefix,
            ftp_file,
            gcp_bucket,
            gcp_prefix,
            gcp_file,
        )
        return
    with FTP(ftp_server) as ftp:
        ftp.login()
        bucket = storage.Client().bucket(gcp_bucket)
        gcp_path = f"{gcp_prefix}/{gcp_file}" if gcp_prefix else gcp_file
        blob = bucket.blob(gcp_path)
        logger.info("Changing directory to %s.", ftp_prefix)
        ftp.cwd(ftp_prefix)
        dir_match = re.match(r"^.*(?P<release_date>\d{4}\/\d{2}\/\d{2}){1}$", ftp.pwd())
        if dir_match:
            logger.info("Found release date!: %s", dir_match.group("release_date"))
        buffer = io.BytesIO()
        logger.info(
            "Retrieving data from: ftp://%s/%s/%s.", ftp_server, ftp_prefix, ftp_file
        )
        ftp.retrbinary(f"RETR {ftp_file}", lambda x: buffer.write(x))
        content = buffer.getvalue().decode("utf-8")
        buffer.close()
        logger.info("Uploading data to: gs://%s/%s.", gcp_bucket, gcp_path)
        blob.upload_from_string("".join(content))


__all__ = ["update_gwas_curation_metadata_command"]



"""CLI tests."""

import pytest
from click.testing import CliRunner
from gentroutils import cli


@pytest.mark.integration_test
def test_help():
    """Test --help flag."""
    runner = CliRunner()
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert "Gentroutils Command Line Interface." in result.output


@pytest.mark.integration_test
def test_run_without_command():
    """Test run without flags and commands."""
    runner = CliRunner()
    result = runner.invoke(cli)
    assert "Gentroutils Command Line Interface." in result.output



"""Cli for gentroutils."""

from __future__ import annotations

import logging
import time

import click
import pyfiglet

from gentroutils.commands.update_gwas_curation_metadata import (
    update_gwas_curation_metadata_command,
)
from gentroutils.commands.utils import set_log_file, set_log_lvl, teardown_cli

logger = logging.getLogger("gentroutils")
logger.setLevel(logging.DEBUG)


@click.group()
@click.option("-d", "--dry-run", is_flag=True, default=False)
@click.option(
    "-v",
    count=True,
    default=0,
    callback=set_log_lvl,
    help="Increase verbosity of the logging. Can be used multiple times. The default log level is ERROR, -v is INFO, -vv is DEBUG",
)
@click.option("-q", "--log-file", callback=set_log_file, required=False)
@click.pass_context
def cli(ctx: click.Context, **kwargs: dict[str, str]) -> None:
    r"""Gentroutils Command Line Interface."""
    ascii_art = pyfiglet.Figlet(font="serifcap").renderText("Gentroutils")
    click.echo(click.style(ascii_art, fg="blue"))
    ctx.max_content_width = 200
    ctx.ensure_object(dict)
    ctx.obj["dry_run"] = kwargs["dry_run"]
    ctx.obj["execution_start"] = time.time()
    ctx.call_on_close(lambda: teardown_cli(ctx))


cli.add_command(update_gwas_curation_metadata_command)

__all__ = ["cli"]



"""CLI submodules for gentroutils package."""

from gentroutils.commands.update_gwas_curation_metadata import (
    update_gwas_curation_metadata_command,
)

__all__ = ["update_gwas_curation_metadata_command"]
